#
# This module attempts to work out the types of each reference.
#

import compiler
import constants
from utils import fqn_rcar, fqn_rcdr, fqn_car, fqn_cdr

import bike.parsing.extended_ast

# parentnode is for when this is called on a sourcenode half way down
# the tree (for reloading source files)
def addtypeinfo(ast,parentnode=None,**opts):
    progressCallback = None
    if opts.has_key('progressCallback'):
        progressCallback = opts['progressCallback']
    compiler.walk(ast,PassOne_ModulesClassesAndFunctions(parentnode))
    compiler.walk(ast,PassTwo_AssignmentsAndReturnsFromFunctionCalls(progressCallback))


class TypeNotSupportedException:
    def __init__(self,msg):
        self.msg = msg

    def __str__(self):
        return self.msg

class TypeInfo:
    def __init__(self,parentScope=None):
        self.typehash = {}
        self.parentScope = parentScope
    def addType(self,name,node):
        self.typehash[name] = node
    def removeType(self,name):
        del self.typehash[name]
        
    def addReturnType(self,type):
        self.typehash['return'] = type

        
    def getTypeOf(self,name):
        return self._getTypeOf_Impl(name)
    
    def getTypeOf_doesntThrowException(self,name):
        try:
            return self._getTypeOf_Impl(name)
        except KeyError:
            return None

    def _getTypeOf_Impl(self,name):
        typeof = None
        if "." in name:   # i.e. fqn
            typeof = self.typehash[fqn_car(name)].typeinfo.getTypeOf(fqn_cdr(name))
        else:
            typeof = self.typehash[name]
        if type(typeof) == type(lambda :1):  # is a closure
            #print "evaluating type"
            typeof = typeof()   # evaluate the closure
        return typeof

    def getReturnType(self):
        return self.getTypeOf_doesntThrowException('return')

    def haskey(self,name):
        return self.typehash.has_key(name)

    def __str__(self):
        mystr = ""
        for key in self.typehash.keys():
            mystr += key + ":" + str(self.getTypeOf_doesntThrowException(key)) + "\n"
        return mystr
        #return str(self.typehash)


    # returns None if can't deduce type of expression.
    def getTypeOfExpr(self,ast):
        if isinstance(ast,compiler.ast.Name):
            try:
                return self.getTypeOf(ast.name)
            except KeyError:
                if self.parentScope is not None:
                    return self.parentScope.typeinfo.getTypeOfExpr(ast)
        elif isinstance(ast,compiler.ast.Getattr):
            expr = self.getTypeOfExpr(ast.expr)
            if expr is not None:
                if isinstance(expr,bike.parsing.extended_ast.Instance):
                    return expr.getType().typeinfo.getTypeOf_doesntThrowException(ast.attrname)
                else:                    
                    return expr.typeinfo.getTypeOf_doesntThrowException(ast.attrname)
            return None
        else:
            raise TypeNotSupportedException, "Evaluation of "+str(ast)+" not supported"

class ClassTypeInfo(TypeInfo):
    def __init__(self,parentScope):
        TypeInfo.__init__(self,parentScope)
        self.subclasses=[]

    def addSubClass(self,subclass):
        self.subclasses.append(subclass)

    def getSubClasses(self):
        return self.subclasses


class ModuleTypeInfo(TypeInfo):
    def __init__(self,parentScope):
        TypeInfo.__init__(self,parentScope)
        self._modulesImportedWithStar = []

    def getTypeOf(self,name):
        type = TypeInfo.getTypeOf_doesntThrowException(self,name)
        if type is None:
            for m in self._modulesImportedWithStar:
                type = m.typeinfo.getTypeOf_doesntThrowException(name)
                if type is not None:
                    break
            if type is None:
                raise KeyError
        return type

    def addModuleImportedWithStar(self,module):
        self._modulesImportedWithStar.append(module)
    

def makeFQN(scopename,leafname):
    if leafname == "":
        return scopename
    elif scopename == "":
        return leafname
    else:
        return scopename+'.'+leafname


class PassOne_ModulesClassesAndFunctions:

    def __init__(self,parentnode):
        self.parentnode = parentnode
        pass
        
    def visitTop(self,node):
        node.typeinfo = TypeInfo()
        for child in node.getChildNodes():
            self.visit(child)

    def visitSource(self,node):
        self.visit(node.module)

    def visitPackage(self,node):
        node.typeinfo = TypeInfo(node.parentScope)
        node.parentScope.typeinfo.addType(node.name,node)
        for child in node.getChildNodes():
            self.visit(child)
        
    def visitModule(self,node):
        node.parentScope.typeinfo.addType(node.name,node)
        node.typeinfo = ModuleTypeInfo(node.parentScope)
        self.visit(node.node)

    def visitClass(self,node):
        node.typeinfo = ClassTypeInfo(node.parentScope)
        node.parentScope.typeinfo.addType(node.name,node)
        for child in node.getChildNodes():
            self.visit(child)

    def visitFunction(self,node):
        node.typeinfo = TypeInfo(node.parentScope)
        node.parentScope.typeinfo.addType(node.name,node)
        if len(node.argnames)!=0 and node.argnames[0] == "self":
            self._addSelfToFunction(node,node.parentScope)
        self.visit(node.code)

    def _addSelfToFunction(self,fnnode,classnode):
        fnnode.typeinfo.addType("self",bike.parsing.extended_ast.Instance(classnode))



class PassTwo_AssignmentsAndReturnsFromFunctionCalls:

    def __init__(self,promptCallback,calldepth=0):
        self.calldepth = calldepth
        self.promptCallback = promptCallback

    def visitPackage(self,node,scope=None):
        for child in node.getChildNodes():
            self.visit(child,node)

    def visitModule(self,node,scope=None):
        if self.promptCallback:
            self.promptCallback("Deducing Type Information",node.name)
        self.visit(node.node,node)

    def visitClass(self,node,scope=None):
        # add this class as subtype to base class
        for basename in node.bases:
            base = scope.typeinfo.getTypeOfExpr(basename)
            if base:
                assert isinstance(base,compiler.ast.Class)
                base.typeinfo.addSubClass(node)                
        self.visit(node.code,node)
    
    def visitFunction(self,node,scope=None):
        self.visit(node.code,node)

    def visitAssign(self,assign,scope):
        if isinstance(assign.expr,compiler.ast.CallFunc):
            # store type as a closure for later (lazy) evaluation
            type = lambda : self._getTypeOfCallFunc(assign.expr,scope)   
            for assnode in assign.nodes:
                if isinstance(assnode,compiler.ast.AssName):
                    scope.typeinfo.addType(assnode.name,type)
                elif isinstance(assnode,compiler.ast.AssAttr):
                    node = scope.typeinfo.getTypeOfExpr(assnode.expr)
                    if node is not None:
                        if isinstance(node,bike.parsing.extended_ast.Instance):
                            node.getType().typeinfo.addType(assnode.attrname,type)
                        else:
                            node.typeinfo.addType(assnode.attrname,type)


    def visitReturn(self,node,scope):
        scope.typeinfo.addReturnType(lambda :self.calcReturnType(node,scope))
        """
        returntype = None
        try:
            returntype = scope.typeinfo.getTypeOfExpr(node.value)
        except TypeNotSupportedException:
            pass
        if returntype is not None:
            scope.typeinfo.addReturnType(returntype)
        elif isinstance(node.value,compiler.ast.CallFunc):
            type = self._getTypeOfCallFunc(node.value,scope)
            if type is not None:
                scope.typeinfo.addReturnType(type)
        """

    def calcReturnType(self,node,scope):
        if isinstance(node.value,compiler.ast.CallFunc):
            return self._getTypeOfCallFunc(node.value,scope)
        else:
            try:
                return scope.typeinfo.getTypeOfExpr(node.value)
            except TypeNotSupportedException:
                return None


    def visitImport(self,node,scope):
        for name,alias in node.names:
            parent_package = fqn_rcdr(scope.fqn)

            if alias is not None:
                importedname = alias
            else:
                if "." in name:
                    # import just the first package, since
                    # getTypeOfExpr() will follow the package typeinfo
                    # nodes to the find the correct type
                    name = name[:name.index('.')]
                importedname = name

            type = lambda name=name:self._getImportedType(scope,name) # closure - lazy evaluation
            scope.typeinfo.addType(importedname,type)


    def visitFrom(self,node,scope):
        if node.names[0][0] == '*': # e.g. from foo import *
            mod = self._getImportedType(scope,node.modname,"*")
            if mod is not None:
                scope.typeinfo.addModuleImportedWithStar(mod)

        else:
            for name,alias in node.names:      
                if alias is not None:
                    importedname = alias
                else:
                    importedname = name
                # closure for lazy eval, just in case module changes (is reloaded)
                #   (name=name insures value rather than ref is used in closure)
                type = lambda name=name:self._getImportedType(scope,node.modname,name)
                scope.typeinfo.addType(importedname,type)


    # returns None if cant get type
    def _getTypeOfCallFunc(self,callfunc,scope):
        try:
            #print "calling ",callfunc
            node = scope.typeinfo.getTypeOfExpr(callfunc.node)
            #print node
        except TypeNotSupportedException:
            return None
        if isinstance(node,compiler.ast.Class):
            return bike.parsing.extended_ast.Instance(node)
        elif isinstance(node,compiler.ast.Function):
            returntype = node.typeinfo.getReturnType()
            # prevent stack overflow on recursive functions
            if returntype is None and node != scope and self.calldepth < constants.MAXCALLDEPTH:
                compiler.walk(node,PassTwo_AssignmentsAndReturnsFromFunctionCalls(self.promptCallback,self.calldepth + 1))
                returntype = node.typeinfo.getReturnType() # could still be None
            return returntype
        else: 
            return None
                                       

    def _getImportedType(self,scope,modname,name=None):
        toptypeinfo = scope.sourcenode.topnode.typeinfo   # code smell?
        
        parent_package = fqn_rcdr(scope.fqn)
        if name == "*" or name == None:
            dot_name = ""
        else:
            dot_name = "."+name
        if toptypeinfo.getTypeOf_doesntThrowException(parent_package + "." +modname + dot_name):
            fqn = parent_package + "." +modname + dot_name
        elif toptypeinfo.getTypeOf_doesntThrowException(modname + dot_name):
            fqn = modname + dot_name
        else:
            return None  # can't find imported type
        return toptypeinfo.getTypeOf_doesntThrowException(fqn)        
        
