from __future__ import nested_scopes
# this module is concerned with turning an ast into a string
import tokenize


def astToString(ast):
    return tokensToString(ast.getTokens())

def tokensToString(tokens):
    if tokens == []:
        return ""
    return TokensToString().tokensToString(tokens)

class TokensToString:

    def tokensToString(self,tokens):
        self.src = ""
        self.tabwidth = 4
        self.indent = 0
        self.startofline = 1
        self.prevToken = tokens[-1]
        self.tokenpos = 0
        self.tokens = tokens
        for token in tokens:
            self.doLineBreakIfRequired(token,self.prevToken)
            self.doIndentIfRequired(token)
            if token.type == tokenize.COMMENT:
                self.handleComment(token)
            elif token.type == tokenize.NL or token.type == tokenize.NEWLINE:
                self.handleNewline()
            elif token.type == tokenize.INDENT:
                self.indent += 1
            elif token.type == tokenize.DEDENT:
                self.indent -= 1
            elif token.type == tokenize.OP:
                self.handleOp(token)
            elif token.type == tokenize.STRING:
                self.handleString(token)
            elif token.type == tokenize.ENDMARKER:
                pass
            else:                
                self.handleName(token)
            self.tokenpos +=1
            self.prevToken = token
        return self.src


    def tokenNeedsIndenting(self,token):
        if token.type == tokenize.COMMENT or \
           token.type == tokenize.STRING or \
           token.type == tokenize.OP or \
           token.type == tokenize.NAME:
            return 1
        return 0


    def handleOp(self,token):
        if self.isAssignment(token) or self.isConditional(token) or self.isStream(token):
            self.src += " "+token.name+" "  # I like assignments to be a = b, not a=b
        elif self.isComma(token):
            self.src += token.name+" "  # I like commas to be a, b, not a,b
        else:
            self.src += token.name
        self.startofline = 0

    def isAssignment(self,token):
        if token.name == "=": return 1
        return 0

    def isStream(self,token):
        if token.name == ">>": return 1
        return 0

    def isComma(self,token):
        if token.name == ",": return 1
        return 0


    def isConditional(self,token):
        if token.name in ["==","!=","<","<=",">",">=","in"]: return 1
        return 0

    def handleString(self,token):
        if self.startofline == 0 and \
           (self.prevToken.type == tokenize.NAME or self.prevToken.name == ":"):
            self.src += " "
        str = self._modifyEscapedCharacters(token.name)
        self.src += str
        self.startofline = 0                

    def _modifyEscapedCharacters(self,str):
        #str = str.replace("\t","\\t")
        #str = str.replace("\n","\\n")
        #str = str.replace("\0","\\0")
        return str
        

    def handleName(self,token):
        if self.startofline == 0 and \
           (self.prevToken.type == tokenize.NAME or self.prevToken.name == ":"):
            self.src += " "

        self.src += token.name
        self.startofline = 0                

    def handleNewline(self):
        self.src += "\n"
        self.startofline = 1

    def handleComment(self,token):
        if not self.startofline:  # inline comment
            for i in range(token.start[1] - self.prevToken.end[1]):
                self.src += " "
        if self.startofline and self.previousTextTokenWasComment():
            currentpos = self.getIndentWidthInSpaces()
            for i in range(token.start[1] - currentpos):
                self.src += " "
        self.src += token.name
        if self.commentIncludesNewline(token):
            self.startofline = 1

    def previousTextTokenWasComment(self):
        if self.prevToken.type == tokenize.COMMENT:
            return 1
        # iterate back through tokens
        pos = self.tokenpos - 1 
        while self.tokens[pos].type == tokenize.INDENT or \
              self.tokens[pos].type == tokenize.NEWLINE:
            pos -= 1    
        return self.tokens[pos].type == tokenize.COMMENT


    def getIndentWidthInSpaces(self):
        return self.indent*self.tabwidth

    def doIndentIfRequired(self,token):
        if self.startofline and self.tokenNeedsIndenting(token):
            if self.tokenpos > 1 and self.tokens[self.tokenpos-2].name == ",":
                for i in range(self._workOutIndentAfterLineBreak(token)+1):
                    self.src += " "
            else:
                self.src += self.getIndentStr()
    

    def doLineBreakIfRequired(self,token,prevtoken):
        if (not self.startofline) and \
           token.type != tokenize.ENDMARKER and \
           token.type != tokenize.NEWLINE and \
           token.start[0] == prevtoken.start[0] + 1:
            self.src += " \\\n"
            indent = self._workOutIndentAfterLineBreak(token)
            for i in range(indent):
                self.src += " "

    
    def _workOutIndentAfterLineBreak(self,token):
        if (self.indent*self.tabwidth) > (token.start[1]-1): # token was probably modified
            return self.getIndentWidthInSpaces()
        else:
            return token.start[1]-1   # probably not modified, so use the x offset

    def getIndentStr(self):
        str=""
        for i in range(self.indent*self.tabwidth):
            str += " "
        return str

    
    def commentIncludesNewline(self,commentToken):
        if commentToken.name.find("\n") == -1:
            return 0
        else:
            return 1



def save(xast,progressCallback=None):
    import compiler
    savedFiles = []
    class SourceVisitor:
        def __init__(self,progressCallback):
            self.progressCallback = progressCallback
        def visitSource(self,node,scope=0):
            if node.isModified():
                if self.progressCallback:
                    self.progressCallback("Saving",node.filename)
                f = open(node.filename,"w+")
                f.write(astToString(node))
                f.close()
                savedFiles.append(node.filename)
                node.setUpToDate()
    compiler.walk(xast,SourceVisitor(progressCallback))
    return savedFiles
