#!/usr/bin/env python
import setpath
import unittest
import compiler
import os

from bike.testutils import *
from bike.mock import Mock

from load import _get_modpkg_path
from load import *


class TestGetFilesForName(unittest.TestCase):
    def testGetFilesForName_recursivelyReturnsFilesInBreadthFirstOrder(self):
        createPackageStructure("pass","pass")
        assert getFilesForName("a") == [os.path.join('a','foo.py'), os.path.join('a','b','bah.py')]
        removePackageStructure()

    def testGetFilesForName_globsStars(self):
        createPackageStructure("pass","pass")
        assert getFilesForName(os.path.normpath("a/fo*")) == [os.path.join('a','foo.py')]
        removePackageStructure()

class TestLoad(unittest.TestCase):
    def testLoad_returnsEmptyTopNodeIfNoFilesMatched(self):
        top = load("doesntexist")
        assert top.modules == []
        assert top.packages == []

        
    def testLoad_returnsASTContainingDepthFirstStructure(self):
        createPackageStructure("pass","pass")
        class ASTVisitor:
            def __init__(self):
                self.visited=[]
            def visitSource(self,node,scope=0):
                self.visited.append(node.filename)
                self.visited.append(node.modulename)
                self.visit(node.module)
            def visitPass(self,node,scope=0):
                self.visited.append('pass')
        top = load("a")
        removePackageStructure()
        a = compiler.walk(top,ASTVisitor())
        assert a.visited == [os.path.join(os.getcwd(),'a','b','bah.py'),'bah','pass',os.path.join(os.getcwd(),'a','foo.py'),'foo','pass']

    def testLoad_returnsTopNodeWithSourceIfOnlyOneFileImported(self):
        writeTmpTestFile("pass")
        top = load(tmpfile)
        class ASTVisitor:
            def __init__(self): self.visited=[]
            def visitSource(self,node,scope=0):
                self.visited.append(node.filename)
                self.visit(node.module)
        a = compiler.walk(top,ASTVisitor())
        assert a.visited == [os.path.abspath(tmpfile)]
        deleteTmpTestFile()

    def testLoad_doesntThrowExceptionCalledTwiceOnDifferentTrees(self):
        createPackageStructure("ClassWithRefToOwnMethod","ClassNestedInFunction")
        top = load("a")
        top = load(os.path.join("a","b"),None,top)

    #if given an multi-element path to root package
    #(e.g. foo/bah/root, doesnt include other directories in package
    #structure
    def testLoad_doesntIncludeDirectoriesPreceedingRootInPackageStructure(self):
        createPackageStructure("pass","pass")
        top = load(os.path.join('a','b','..','..','a'))
        class ASTVisitor:
            def __init__(self): self.visited=[]
            def visitPackage(self,node):
                self.visited.append(node.name)
                for i in node.getChildNodes():
                    self.visit(i)
        a = compiler.walk(top,ASTVisitor())
        assert len(a.visited) == 2
        assert a.visited[0] == 'a'
        assert a.visited[1] == 'b'



class TestLoadUICallback(unittest.TestCase):
    class MockCallback(Mock):
        def uiCallBack(self,action,filename):
            pass    
    """
    def test_callsUICallback(self):
        m = self.MockCallback()
        createPackageStructure("pass","pass")
        load("a",m.uiCallBack)
        removePackageStructure()
        assert len (m.mockCalledMethods["uiCallBack"]) == 3
        assert m.mockCalledMethods["uiCallBack"][0].params[0] == "Parsing"
        assert m.mockCalledMethods["uiCallBack"][0].params[1] == os.path.join("a","foo.py")
        assert m.mockCalledMethods["uiCallBack"][1].params[0] == "Parsing"
        assert m.mockCalledMethods["uiCallBack"][1].params[1] == os.path.join("a","b","bah.py")
        assert m.mockCalledMethods["uiCallBack"][2].params[0] == "Deducing Type Information"
        assert m.mockCalledMethods["uiCallBack"][2].params[1] == ""
    """

if __name__ == "__main__":
    unittest.main()
