#!/usr/bin/env python 
import setpath
import tokenhandler
import tokenutils
from bike.parsing.extended_ast import Top
import compiler.ast 

import unittest
from tokenutils import getTokenList
from test_brmtransformer import getASTNodeFromSrc
from bike.testutils import *
from bike import testdata
import output



class MockNode(compiler.ast.Node):
    def getTokens(self):
        return ["childtoken"]


class TestEmptyTokenHandler(unittest.TestCase):
    class MockParent(tokenhandler.EmptyTokenHandler):
        def getChildNodes(self):
            return [MockNode()]
        
    def testCallsChilden(self):
        node = TestEmptyTokenHandler.MockParent()
        tokens = node.getTokens()
        assert(tokens == ["childtoken"])
        

class TestCompoundNodeTokenHandler:

    class MockParent(tokenhandler.CompoundNodeTokenHandler):
        def getChildren(self):
            return [MockNode()]

    # template method
    def getNodeToTest(self):
        return TestCompoundNodeTokenHandler.MockParent()

    def testGetTokens_CallsChilden(self):
        node = self.getNodeToTest()
        node.addTokens(["beforetoken"],["aftertoken"])
        tokens = node.getTokens()
        assert(tokens == ["beforetoken","childtoken","aftertoken"])



testdata_function_withPreceedingComment="""
# this is a comment
# this is still a comment
def afunction(foo):
    pass
"""

testdata_function = """
def afunction(foo):
    pass
"""


class TestClassTokenHandler(unittest.TestCase):

    def testRename_renamesTheClassAST(self):
        ast = createSourceNodeAt(Top(),testdata.class_and_method,"a")
        addtypeinfo(ast)
        ast = getASTNode(ast,"Class")
        ast.rename("bah")
        assert ast.name == "bah"
        


class TestFunctionTokenHandler(unittest.TestCase):

    def testGetStartCoords_returnsCorrectCoordsForASTOnFirstLine(self):
        src = "def foo(self): pass"
        ast = createSourceNodeAt(Top(),src,"a")
        node = getASTNode(ast,"Function")
        assert node.getStartCoords().x == 0
        assert node.getStartCoords().y == 0
        assert node.getEndCoords().x == 19
        assert node.getEndCoords().y == 0


    def testGetNodeTokens_returnsCorrectTokensForOneLineFunction(self):
        src = "def foo(self): pass"
        ast = createSourceNodeAt(Top(),src,"a")
        node = getASTNode(ast,"Function")
        assert node.getNodeTokens() == getTokenList(src)[:-1]
        
        
    def testGetCoords_returnsCorrectCoordsForASTNodeNotOnFirstLine(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImportAs[1:]
        ast = createSourceNodeAt(Top(),src,"a")
        g = getASTNode(ast,"Function")
        assert g.getStartCoords().x == 0
        assert g.getStartCoords().y == 2
        assert g.getEndCoords().x == 0
        assert g.getEndCoords().y == 5


    def testRename_renamesTheFunction(self):
        ast = createSourceNodeAt(Top(),testdata_function,"a")
        addtypeinfo(ast)
        ast = getASTNode(ast,"Function")
        ast.rename("bah")
        assert ast.name == "bah"
            
    def testRename_updatesTheTokens(self):
        ast = createSourceNodeAt(Top(),testdata_function,"a")
        addtypeinfo(ast)
        ast = getASTNode(ast,"Function")
        ast.rename("bah")
        tokens = ast.getTokens()
        assert tokens[1].name == "bah"
        
    def testRename_renamesTheTypeinfoFQN(self):
        ast = createSourceNodeAt(Top(),testdata.class_and_method,"a")
        addtypeinfo(ast)
        ast = getASTNode(ast,"Function")
        ast.rename("bah")
        assert ast.fqn == "a.a.bah"


    def testRename_updatesTheTokensIfPreceedingComments(self):
        ast = createSourceNodeAt(Top(),testdata_function_withPreceedingComment,"a")
        addtypeinfo(ast)
        ast = getASTNode(ast,"Function")
        ast.rename("bah")
        tokens = ast.getTokens()
        assert(tokens[3].name == "bah")

    def testRename_setsSourceToModified(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImport
        ast = createSourceNodeAt(Top(),src,"a")
        addtypeinfo(ast)
        f = getASTNode(ast,"Function")
        s = getASTNode(ast,"Source")
        f.rename("newName")
        assert s.isModified()



    def testGetNodeTokens_returnsTokensStartingWithDefToken(self):
        ast = getASTNodeFromSrc(testdata.a_class_with_two_methods,"Function",1)
        assert ast.getNodeTokens()[0].name == "def"

class TestGetattrTokenHandler(unittest.TestCase):
    def testRenameAttr_renamesTheAttributeNameInAST(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImport
        ast = createSourceNodeAt(Top(),src,"a")
        g = getASTNode(ast,"Getattr")
        g.renameAttr("newName")
        assert g.attrname == "newName" 

    def testRenameAttr_renamesTheAttributeNameInTokenlist(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImport[1:]
        ast = createSourceNodeAt(Top(),src,"a")
        g = getASTNode(ast,"Getattr")
        g.renameAttr("newName")
        assert g.getNodeTokens()[4].name == "newName" 

    def testRenameAttr_renamesTheAttributeNameInTokenlistWhenTokenNotSecondTokenOfList(self):
        src = testdata.RefCalledOnClassInstanceCreation
        ast = createSourceNodeAt(Top(),src,"a")
        g = getASTNode(ast,"Getattr")
        g.renameAttr("newName")
        assert g.getNodeTokens()[4].name == "newName" 

    def testRenameAttr_setsSourceToModified(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImport
        ast = createSourceNodeAt(Top(),src,"a")
        g = getASTNode(ast,"Getattr")
        s = getASTNode(ast,"Source")
        g.renameAttr("newName")
        assert s.isModified()

    def testGetStartCoords_returnsCorrectCoordsForASTOnFirstLine(self):
        src = "a.b"
        ast = createSourceNodeAt(Top(),src,"a")
        g = getASTNode(ast,"Getattr")
        assert g.getStartCoords().x == 0
        assert g.getStartCoords().y == 0
        assert g.getEndCoords().x == 3
        assert g.getEndCoords().y == 0

    def testGetCoords_returnsCorrectCoordsForASTNodeNotOnFirstLine(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImportAs[1:]
        ast = createSourceNodeAt(Top(),src,"a")
        g = getASTNode(ast,"Getattr")
        assert g.getStartCoords().x == 4
        assert g.getStartCoords().y == 4
        assert g.getEndCoords().x == 15
        assert g.getEndCoords().y == 4


class TestImportTokenHandler(unittest.TestCase):
    def test_renamesTokensWhenFirstImportHasSameClassNameAsSecond(self):
        src = "import bah.foo, baz.foo"
        ast = createSourceNodeAt(Top(),src,"a")
        f = getASTNode(ast,"Import")
        f.rename(1,"bobbins")
        assert output.astToString(f) == "import bah.foo, baz.bobbins"
        

    def test_renamesClassNameTokenWhenClassIsSameNameAsModule(self):
        src = "import bah.bah"
        ast = createSourceNodeAt(Top(),src,"a")
        f = getASTNode(ast,"Import")
        f.rename(0,"bobbins")
        assert output.astToString(f) == "import bah.bobbins"

    def test_renamesClassNameTokenWhenCommentAtEndOfImport(self):
        src = "import bah.bah  # comment"
        ast = createSourceNodeAt(Top(),src,"a")
        f = getASTNode(ast,"Import")
        f.rename(0,"bobbins")
        assert output.astToString(f) == "import bah.bobbins  # comment"

class TestFromTokenHandler(unittest.TestCase):
    def test_renamesTokensWhenModuleIsSameNameAsImportedClass(self):
        src = "from baz import foo, baz"
        ast = createSourceNodeAt(Top(),src,"a")
        f = getASTNode(ast,"From")
        f.rename(1,"bah")
        assert output.astToString(f) == "from baz import foo, bah"



if __name__ == "__main__":
    unittest.main()
