"""
This module contains base classes which hold and manipulate token
state for ast Nodes.
The ast nodes subclass these classes by means of hackery in the extended_ast.py module

Author Phil Dawes (pdawes@users.sourceforge.net)
"""

from compiler.ast import Node
import tokenutils
import tokenize
from utils import fqn_rcdr, fqn_rcar

class coords:
    def __init__(self,x,y):
        self.x = x
        self.y = y

class EmptyTokenHandler:
    def getTokens(self):
        tokens = []
        c = self.getChildNodes()
        for i in c:
            if isinstance(i, Node):
                tokens+=i.getTokens()
        return tokens

class SimpleTokenHandler:
    def getTokens(self):
        if hasattr(self,"sourcenode"):   # this is a hack to allow old tests to work
            self.sourcenode.tokenize()
        return self.getTokens_impl()

    def addTokens(self, tokens):
        self.tokens = tokens

    def getNodeTokens(self):
        return self.getTokens()

    def getTokens_impl(self):
        if hasattr(self,"tokens"):
            return self.tokens
        else:
            return []

    def getStartCoords(self):
        tokens = self.getNodeTokens()
        return coords(tokens[0].start[1],tokens[0].start[0]-1)
        
    def getEndCoords(self):
        tokens = self.getNodeTokens()
        return coords(tokens[-1].end[1],tokens[-1].end[0]-1)


class CompoundNodeTokenHandler(SimpleTokenHandler):

    def addTokens(self, tokensBefore, tokensAfter = []):
        self.tokensBefore = tokensBefore
        self.tokensAfter = tokensAfter

    def getTokens_impl(self):
        tokens = []
        tokens+=self.getTokensBefore()
        c = self.getChildren()
        for i in c:
            if isinstance(i, Node):
                tokens+=i.getTokens()
        tokens+=self.getTokensAfter()
        return tokens

    def getTokensBefore(self):
        return self.tokensBefore

    def getTokensAfter(self):
        return self.tokensAfter

    def getNodeTokens(self):
        """ abstract method """
class ModuleTokenHandler(CompoundNodeTokenHandler):
    def getTokensWithoutPreceedingComments(self):
        pos = 0
        tokens = self.getTokens()
        while(tokens[pos].type == tokenize.COMMENT):
            pos+=1
        return tokens[pos:]

    def getNodeTokens(self):
        return self.getTokensWithoutPreceedingComments()


class ClassTokenHandler(CompoundNodeTokenHandler):

    def getNodeTokens(self):
        pos = 0
        tokens = self.getTokens()
        def isClassToken(token):
            return token.type == tokenize.NAME and token.name == "class"
        while not isClassToken(tokens[pos]):
            pos+=1
        return tokens[pos:]

    def addTokens(self, tokensBefore, tokensMid, tokensAfter):
        self.tokensBefore = tokensBefore
        self.tokensMid = tokensMid
        self.tokensAfter = tokensAfter

    def getTokensMid(self):
        return self.tokensMid

    def getTokens_impl(self):
        tokens = []
        tokens+=self.getTokensBefore()

        if self.bases != []:
            for i in self.bases:
                if isinstance(i, Node):
                    tokens+=i.getTokens()
            tokens += self.getTokensMid()
            
        tokens+=self.code.getTokens()
        tokens+=self.getTokensAfter()
        return tokens



    def rename(self,name):
        self.name = name
        # update the token
        tokens = self.getNodeTokens()
        tokens[1].name = name
        self.fqn = self.parentScope.fqn+"."+name
        #update the typeinfo
        self.sourcenode.topnode.typeinfo.addType(self.fqn,self)        
        self.parentScope.typeinfo.addType(name,self)        
        self.sourcenode.setModified()
        
        

# methods which update the Function and keep tokens in sync
class FunctionTokenHandler(CompoundNodeTokenHandler):

    def getNodeTokens(self):
        pos = 0
        tokens = self.getTokens()
        def isDefToken(token):
            return token.type == tokenize.NAME and token.name == "def"
        while not isDefToken(tokens[pos]):
            pos+=1
        return tokens[pos:]


    def rename(self, name):
        self.sourcenode.tokenize()
        oldname = self.name
        self.name = name
        # update the token
        tokens = self.getNodeTokens()
        tokens[1].name = name
        oldfqn = self.fqn
        newfqn = self.parentScope.fqn+"."+name
        self.fqn = newfqn
        #update the typeinfo                
        self.sourcenode.topnode.typeinfo.addType(self.fqn,self)        
        self.parentScope.typeinfo.addType(name,self)
        self.sourcenode.setModified()




class GetattrTokenHandler(CompoundNodeTokenHandler):
    def getNodeTokens(self):
        tokens = []
        for child in self.getChildNodes():
            tokens+=child.getTokens()
            tokens+=self.getTokensAfter()
        return tokens

    def renameAttr(self, newname):
        for token in self.getNodeTokens():
            if token.name == self.attrname:
                token.name = newname
        self.attrname = newname
        self.sourcenode.setModified()

class NameTokenHandler(SimpleTokenHandler):
    def rename(self,newname):
        nodetokens = self.getNodeTokens()
        nodetokens[0].name = newname
        oldname = self.name
        self.name = newname
        if self.parentScope.typeinfo.haskey(oldname):
            type = self.parentScope.typeinfo.getTypeOf(oldname)
            self.parentScope.typeinfo.addType(newname,type)
        self.sourcenode.setModified()
        
    def getNodeTokens(self):
        if not hasattr(self,"tokens"):
            return []
        pos = 0
        def isNameToken(token):
            return token.type == tokenize.NAME and token.name == self.name
        while not isNameToken(self.tokens[pos]):
            pos+=1
        return self.tokens[pos:]
        
class ImportTokenHandler(SimpleTokenHandler):
    def getNodeTokens(self):
        pos = 0
        tokens = self.getTokens()
        def isImportToken(token):
            return token.type == tokenize.NAME and token.name == "import"
        while not isImportToken(tokens[pos]):
            pos+=1
        return tokens[pos:]


    def rename(self,indexOfImport,newname):
        name = self.names[indexOfImport]
        oldfqn = name[0]
        oldname = fqn_rcar(oldfqn)
        self.names[indexOfImport] = (fqn_rcdr(oldfqn) + "."+newname,name[1])
        nodetokens = self.getNodeTokens()
        tokens = self._getTokensCorrespondingToImportName(indexOfImport)
        tokens[-1].name = newname
        self.sourcenode.setModified()

    def _getIndexOfFirstTokenCorrespondingToImportName(self,indexOfImport):
        tokens = self.getNodeTokens()[1:]
        commacount = 0
        for i in range(len(tokens)):
            if commacount == indexOfImport:
                return i
            if tokens[i].name == ",":
                commacount += 1

    # needs refactoring
    def _getTokensCorrespondingToImportName(self,indexOfImport):
        tokens = self.getNodeTokens()[1:]
        commacount = 0
        starttoken = 0
        endtoken = 0
        for i in range(len(tokens)):
            if commacount == indexOfImport:
                starttoken = i
                break
            if tokens[i].name == ",":
                commacount += 1
        def isAnImportToken(token):
            return token.name == "." or token.type == tokenize.NAME
        for i in range(starttoken,len(tokens)):
            if not isAnImportToken(tokens[i]):
                endtoken = i
                break
        if endtoken == 0:
            endtoken = len(tokens)            
        return tokens[starttoken:endtoken]

        


class FromTokenHandler(SimpleTokenHandler):
    def getNodeTokens(self):
        pos = 0
        tokens = self.getTokens()
        def isFromToken(token):
            return token.type == tokenize.NAME and token.name == "from"
        while not isFromToken(tokens[pos]):
            pos+=1
        return tokens[pos:]


    def getImportTokens(self):
        pos = 0
        tokens = self.getTokens()
        def isImportToken(token):
            return token.type == tokenize.NAME and token.name == "import"
        while not isImportToken(tokens[pos]):
            pos+=1
        return tokens[pos:]


    def rename(self,indexOfImport,newname):
        name = self.names[indexOfImport]
        oldname = name[0]
        oldfqn = self.modname + "." + oldname
        self.names[indexOfImport] = (newname,name[1])
        nodetokens = self.getImportTokens()
        tokidx = self._getIndexOfTokenCorrespondingToImportName(nodetokens,
                                                                indexOfImport)
        nodetokens[tokidx].name = newname
        self.sourcenode.setModified()

    def _getIndexOfTokenCorrespondingToImportName(self,tokens,indexOfImport):
        commacount = 0
        for i in range(1,len(tokens)):
            if commacount == indexOfImport:
                return i
            if tokens[i].name == ",":
                commacount += 1
