# Utilities to handle matching of AST nodes and parse trees to the
# scanner tokens they represent. Useful because BRM needs to convert
# AST nodes back to source, and this is easier to do if you have the
# corresponding scanner tokens.

from __future__ import nested_scopes

import tokenize
import token
import types
import StringIO

def getTokenList(str):
    return TokenList().processString(str)

def getTokenListFromFile(path):
    return TokenList().processFile(path)

def getTokenConverter(str):
    return TokenConverter(getTokenList(str))


# produces a list of tokens
# use   t = TokenList("filename.py").getList()
#
# N.B. this class does some token reordering to ensure that comments
# immediately before a block of code are placed after the previous
# block DEDENT token
class TokenList:

    def __init__(self):
        self.list = []
        self.tmpComment = []
        self.cacheComments = 0

    def processString(self,str):        
        tokenize.tokenize(StringIO.StringIO(str).readline,self.process_token)
        self.runTransformers()
        return self.list

    def processFile(self,path):
        tokenize.tokenize(open(path).readline,self.process_token)
        self.runTransformers()
        return self.list

    def runTransformers(self):
        self.putDedentTokenBeforeCommentsDescribingNextBlock()
        self.putIndentTokenBeforeCommentsDescribingBlock()
        pass

    def process_token(self,ttype,tname,tstart,tend,tline):
        t = TokenList.token(ttype,tname,tstart,tend,tline)
        self.list.append(t)


    def putDedentTokenBeforeCommentsDescribingNextBlock(self):        
        def indexOfPreviousNLCommentBlock(i):
            j=i
            while self.list[j-1].type == tokenize.COMMENT: j-=1
            while self.list[j-1].type == tokenize.NL: j-=1
            if self.list[j].type == tokenize.NL:
                return j
            else: return i
            
        for i in range(0,len(self.list)):
            if self.list[i].type == tokenize.DEDENT:

                j=i
                while indexOfPreviousNLCommentBlock(j) != j:
                    j=indexOfPreviousNLCommentBlock(j)
                                    
                # move the dedent to before the comment
                dedenttok = self.list[i]
                del self.list[i]
                self.list.insert(j,dedenttok)
                

    # ensure comments describing block are indented at the same level
    def putIndentTokenBeforeCommentsDescribingBlock(self):
        def getIndexOfTokenAfterPreviousNewline(i):
            while self.list[i-1].type != tokenize.NEWLINE:
                i-=1
            return i
            
        
        for i in range(0,len(self.list)):
            if self.list[i].type == tokenize.INDENT:
                j = getIndexOfTokenAfterPreviousNewline(i)
                if j != i:                    
                    indenttok = self.list[i]
                    del self.list[i]
                    self.list.insert(j,indenttok)

    
    class token:
        def __init__(self,type,name,start,end,line):
            self.type = type
            self.name = name
            self.start= start
            self.end = end
            self.line = line

        def __eq__(self,other):
            if (self.type != other.type): return 0
            if (self.name != other.name): return 0
            return 1

        def __ne__(self,other):
            return  not self.__eq__(other)

        def __repr__(self):
            if self.tokenDoesntRequireName():
                name = ""
            else: name = `self.name`
            return token.tok_name[self.type]+name+`self.start`

        
        def tokenDoesntRequireName(self):
            return self.type == token.NEWLINE\
                      or self.type == token.INDENT \
                      or self.type == token.DEDENT \
                      or self.type == tokenize.NL
                

# iterates through a list of tokens, providing utilities to match
# sections of parse tree to the corresponding scanner tokens in order.
class TokenConverter:
    def __init__(self,tokenlist):
        self.tokens = tokenlist
        self.pos = 0

    def hasMoreTokens(self):
        return len(self.tokens) > self.pos
            

    def peekNextToken(self):
        try:
            t = self.tokens[self.pos]
        except:
            t = 0 
        return t

    def popToken(self):
        self.pos +=1
        return self.tokens[self.pos-1]
        
    def getTokens(self,parsetree):
        tokens = []
        self._getTokensRecurse(parsetree,tokens)
        return tokens
    
    def getTokensPreceedingNodes(self,node):
        pos = self.pos
        self.pos = self._getOffsetOfTokensCorrespondingToParseTree(node)
        return self.tokens[pos:self.pos]

    def getTokensTo(self,tokenname):
        t=[]
        while(self.peekNextToken().name != tokenname):
            t.append(self.popToken())
        t.append(self.popToken()) # get the target token
        return t

    def getTokensStartBlock(self):
        t = self.getTokensTo(":")
        # if next token is a COMMENT, include it
        if self.peekNextToken().type == tokenize.COMMENT:
            t.append(self.popToken())
        # if next token is a NEWLINE, include it
        if self.peekNextToken().type == tokenize.NEWLINE:
            t.append(self.popToken())
        # if next token is an INDENT, include it
        if self.peekNextToken().type == tokenize.INDENT:
            t.append(self.popToken())
        # if next token is an STRING (docstring), include it
        if self.peekNextToken().type == tokenize.STRING:
            t.append(self.popToken())
            # and the following NEWLINE
            if self.peekNextToken().type == tokenize.NEWLINE:
                t.append(self.popToken())
        return t

    def getTokensIncWhitespace(self,parsetree):
        t = self.getTokens(parsetree)
        t += self.getWhitespaceLines()
        return t

    def getDedent(self):
        t = self.getDedentIfPresent()
        if t == []:
            raise "next token wasn't a dedent"
        return t

    def getDedentIfPresent(self):
        if self.peekNextToken().type == tokenize.DEDENT:
            t = [self.popToken()] # pop the DEDENT
            t += self.getWhitespaceLines()
            return t
        else:
            return []
    
    def getWhitespaceLines(self):
        t = []
        try:
            if self.peekNextToken().type == tokenize.NEWLINE:
                t.append(self.popToken())
            # should never get 2 newlines in a row
            assert(self.peekNextToken().type != tokenize.NEWLINE)
            while self.peekNextToken().type == tokenize.NL:
                t.append(self.popToken())
        except:
            pass
        return t


    def getTokensUptoLastNewlineBeforeCode(self):
        tokens = []
        pos = self.pos
        while(self.tokens[pos].type == tokenize.COMMENT or
              self.tokens[pos].type == tokenize.NL):
            if(self.tokens[pos].type == tokenize.NL):
                self._appendTokens(self.tokens[pos].type,
                                      self.tokens[pos].name,tokens)
            pos += 1
        return tokens
        

    def getRestOfTokens(self):
        tokens = self.tokens[self.pos:]
        self.pos = len(self.tokens)
        return tokens


    # returns the collection of tokens making up the comment.
    #  (has heuristics to judge tokens when comment is multiline)
    # returns empty list if next token not comment
    def getCommentAndNewline(self):

        def nextTokenIsPartOfSameComment(commenttoken):
            tok = self.peekNextToken()
            # if they're tabbed to the same place
            return tok and tok.type == tokenize.COMMENT and \
                   commenttoken.start[1] == tok.start[1]

        def getNewlineIfPresent():
            if self.peekNextToken().type == tokenize.NEWLINE:
                return [self.popToken()]
            else: return []


        if self.peekNextToken().type == tokenize.NEWLINE:
            return [self.popToken()] # no comment, so return newline on its own

        tokens = []
        
        if self.peekNextToken().type == tokenize.COMMENT:
            comment = self.popToken()
            tokens.append(comment)
            tokens += getNewlineIfPresent()
            while nextTokenIsPartOfSameComment(comment):
                comment = self.popToken()
                tokens.append(comment)
        return tokens



    ############################################################
    #private methods

    def _getListSlice(self,first,last):
        return self.tokens[first:last+1]
        

    def _getOffsetOfTokensCorrespondingToParseTree(self,parsetree):
        tokennodes = self._getTokenNodesExceptNewlines(parsetree)

        def enoughTokensLeftToMatchNodes(nodes,pos):
            return len(self.tokens) - pos >= len(nodes)

        def nodesMatchTokens(nodes,pos):
            if not enoughTokensLeftToMatchNodes(nodes,pos):
                raise IndexError, "not enough tokens left to match nodes:"+str(nodes)+"\nTokens left:"+str(self.tokens[pos:])

            for i in nodes:
                if self._isMatchingToken(self.tokens[pos],i[0],i[1]) == 0:
                    # NLs can appear in method calls without
                    # '\'. These arent captured in a parsetree so be
                    # ready for the superflous NL token
                    if self.tokens[pos].type == tokenize.NL:
                        if (self._isMatchingToken(self.tokens[pos+1],i[0],i[1]) == 0):
                            return 0
                        else:
                            pos += 2
                    else:
                        return 0
                else:
                    pos += 1
            return 1

        pos = self.pos
        while nodesMatchTokens(tokennodes,pos) == 0:
            pos = pos+1
        return pos


    

    # iterate through tree of nodes looking for the first one representing
    # a token
    def _getFirstTokenNode(self,node):
        if type(node) == types.ListType or type(node) == types.TupleType:
            if self._nodeIsToken(node):
                return node
            else:
                for i in node:     
                    res = self._getFirstTokenNode(i)
                    if res != 0:  # i.e. found it!
                        return res 
        return 0

    # returns a list of nodes that correspond to tokens
    def _getTokenNodes(self,node):

        def getTokenNodesRecursive(node,tokennodes):
            if type(node) == types.ListType or type(node) == types.TupleType:
                if self._nodeIsToken(node):
                    tokennodes.append(node)
                else:
                    for i in node:
                        getTokenNodesRecursive(i,tokennodes)
        tokennodes = []
        getTokenNodesRecursive(node,tokennodes)
        return tokennodes
        

    # returns a list of nodes that correspond to tokens
    # (but don't include newlines or endmarkers)
    def _getTokenNodesExceptNewlines(self,node):

        def getTokenNodesRecursive(node,tokennodes):
            if type(node) == types.ListType or type(node) == types.TupleType:
                if self._nodeIsToken(node) and \
                   node[0] != tokenize.NEWLINE and \
                   node[0] != tokenize.ENDMARKER:
                    tokennodes.append(node)
                else:
                    for i in node:
                        getTokenNodesRecursive(i,tokennodes)
        tokennodes = []
        getTokenNodesRecursive(node,tokennodes)
        return tokennodes



    def _getTokensRecurse(self,node,tokens):
        if type(node) == types.ListType or type(node) == types.TupleType:
            # possible token?
            if self._nodeIsToken(node):
                self._appendTokens(node[0],node[1],tokens)            
            else:
                for i in node:
                    self._getTokensRecurse(i,tokens)

                    
    def _nodeIsToken(self,node):
        # possible token?
        if len(node) == 3 and \
           type(node[0])==types.IntType and \
           type(node[1]) == types.StringType:
            try:
                a = token.tok_name[node[0]]
                return 1   # if above doesn't fail
            except:
                return 0
        return 0
        


    # get tokens preceeding the one with given type/name
    def _getPreceedingTokens(self,type,name):
        before_pos = self.pos
        if len(self.tokens) == self.pos:
            print "warning - reached end of token stream"
            return
        while 1:
            t = self.peekNextToken()
            if self._isMatchingToken(t,type,name):
                break
            else:
                self.popToken()  #not the target token, so pop it
        return self.tokens[before_pos:self.pos]
        

    def _isMatchingToken(self,token,nodetype,nodename):
        if nodetype == tokenize.NAME:
            if nodetype == token.type and nodename == token.name:
                return 1        
        if token.type == tokenize.OP and nodename == token.name:
            return 1
        if token.type == nodetype and token.type != tokenize.NAME:
            return 1
        return 0
        

    # append tokens until you find the node
    def _appendTokens(self, nodetype, nodename, tokens):
        #print "nodename is",nodename
        before_pos = self.pos
        if len(self.tokens) == self.pos:
            print "warning - reached end of token stream"
            return
        while 1:
            #print "token is",self.tokens[self.pos]
            t = self.tokens[self.pos]
            self.pos += 1
            if nodetype == token.NAME and t.type == token.NAME and nodename == t.name:
                break
            if t.type == token.OP and nodename == t.name:
                break
            if t.type != token.OP and t.type != token.NAME and t.type == nodetype:
                break
            # happens when source is all on one line -
            # parstree says newline, tokenstream says endmarker
            if t.type == token.ENDMARKER and nodetype == token.NEWLINE:
                break
            if len(self.tokens) == self.pos:
                print "warning - reached end of token stream"
                break
        tokens += self.tokens[before_pos:self.pos]

