#!/usr/bin/env python
from __future__ import nested_scopes
from bikecli import cli
import unittest
import os


from bike.parsing.tokenutils import getTokenList
from bike.parsing.brmtransformer import parse

from bike import testdata

from bike.testutils import *

from bike.refactor.test_renameMethod import RenameMethodTests, RenameMethodReferenceTests_ImportsClass, RenameMethodReferenceTests_doesntImportClass, RenameMethodAfterPromptTests
from bike.parsing.addtokens import addTokens


# Runs rename method cli against files on the disk, then
# parses them and asserts
class TestCLIRenameMethodWithSingleFile(RenameMethodTests, RenameMethodReferenceTests_doesntImportClass, unittest.TestCase):

    # template method
    def runTarget(self,src,klass,method,newname):
        writeTmpTestFile(src)
        cli().doRenameMethod(tmpfile,tmpmodule+"."+klass,method,newname)
        out = readTmpTestFile()
        deleteTmpTestFile()
        ast = parse(out)
        addTokens(ast,out)
        return ast

class TestCLIRenameMethodReferenceWithDirectoryStructure(RenameMethodTests, RenameMethodReferenceTests_doesntImportClass, unittest.TestCase):

    # template method
    def runTarget(self,src,klass,method,newname):
        createPackageStructure("pass",src)
        cli().doRenameMethod("a","a.b.bah."+klass,method,newname)
        newsrc = readFile(os.path.join('a','b','bah.py'))
        ast = parse(newsrc)
        addTokens(ast,newsrc)
        removePackageStructure()
        return ast

class MockIO(Mock):
    def __init__(self):
        Mock.__init__(self)
        self.str = ""
    def write(self,output):
        self.str += output
    def read(self):
        return "y"

class TestRenameMethodPromptsForUnknownTypes(unittest.TestCase):

    # template method
    def runTarget(self,src,klass,method,newname,io):
        createPackageStructure("pass",src)
        cli(io).doRenameMethod("a","a.b.bah."+klass,method,newname)
        ast = parse(readFile(os.path.join('a','b','bah.py')))
        removePackageStructure()
        return ast
    
    def testRenameMethod_doesntPromptTheUserIfCanDeduceAllTheTypes(self):
        src = testdata.ObjectReturnedByFunctionAndUsedByCallerInPreviousFunction[1:]
        io = MockIO()
        ast = self.runTarget(src,"TheClass","theMethod","newName",io)
        assert len(io.mockCalledMethods) == 0  # not prompted

    def test_doesntPromptTheUserIfCanDeduceAllTheTypesAndMethodRefIsNotOnAnInstanceOfTheTargetClass(self):
        src = testdata.ObjectReturnedByFunctionAndUsedByCallerInPreviousFunction[1:]
        io = MockIO()
        ast = self.runTarget(src,"NotTheClass","theMethod","newName",io)
        assert len(io.mockCalledMethods) == 0  # not prompted

    def testRenameMethod_promptsTheUserIfCantDeduceAllTheTypes(self):
        src = testdata.ObjectRefWithNoDeducableType[1:]
        io = MockIO()
        ast = self.runTarget(src,"TheClass","theMethod","newName",io)
        assert len(io.mockCalledMethods["read"]) == 1
        assert io.str == shallIRenamePrompt[1:-1]

class TestRenameMethodTellsYouAboutChanges(unittest.TestCase):

    def test_tellsUserIfMethodsWereRenamed(self):
        src = testdata.class_and_method
        io = MockIO()
        writeTmpTestFile(src)
        cli(io).doInteractiveRenameMethod(tmpfile,tmpmodule+".a","foo","bah")
        deleteTmpTestFile()
        assert io.str == oneMethodRenamed

    """
    def test_tellsUserIfMethodReferencesWereRenamed(self):
        src = testdata.ObjectCreationAndMethodReference
        io = MockIO()
        writeTmpTestFile(src)
        cli(io).doInteractiveRenameMethod(tmpfile,tmpmodule+".TheClass","theMethod","bah")
        deleteTmpTestFile()
        print io.str
        assert io.str == oneMethodReferenceRenamed
    """

    def test_includesAdviceIfNothingWasChanged(self):
        src = testdata.class_and_method
        io = MockIO()
        writeTmpTestFile(src)
        cli(io).doInteractiveRenameMethod(tmpfile,tmpmodule+".a","notthemethod","bah")
        deleteTmpTestFile()
        assert io.str == noMethodsOrRefsRenamed


class TestRenameMethodRenamesStuffCorrectlyAfterUserPromptedAndSaysYes(RenameMethodAfterPromptTests):
    def runTarget(self,src,klass,method,newname):
        writeTmpTestFile(src)
        cli().doRenameMethod(tmpfile,tmpmodule+"."+klass,method,newname,MockIO())
        out = readTmpTestFile()
        deleteTmpTestFile()
        return parse(out)



class TestRenameMethodDisplaysProgress(unittest.TestCase):
    def test_progressNotDisplayedIfFlagNotPassed(self):
        io = MockIO()
        createPackageStructure(testdata.class_and_method,"pass")
        cli(io,0).doRenameMethod("a","a.foo.a","foo","newname")
        removePackageStructure()
        assert io.str == ""

    def test_progressDisplayedCorrectly(self):
        io = MockIO()
        createPackageStructure(testdata.class_and_method,"pass")
        cli(io,1).doRenameMethod("a","a.foo.a","foo","newname")
        removePackageStructure()
        assert io.str == testdata_progress[1:-1]
        

testdata_progress="""
Parsing..
Deducing Type Information.
Renaming Method.
Renaming Method References..
Saving.
"""


shallIRenamePrompt = """
In scope a.b.bah.foo:
Could not deduce the instance type used in the expression 'a.theMethod'
Context:

def foo():
    a = blah.foobah()
    a.theMethod()

Should I rename this?
"""

oneMethodRenamed = """
1 method was renamed
0 method references were renamed
"""

oneMethodReferenceRenamed = """
0 methods were renamed
1 method reference was renamed
"""

noMethodsOrRefsRenamed = """
0 methods were renamed
0 method references were renamed
Hmmm... nothing was renamed.
Did you use a fully qualified classname (including the root module/package)?
"""




if __name__ == "__main__":
    unittest.main()
