#!/usr/bin/env python
import unittest
import setpath

from bike import testdata
from bike.testutils import*
from bikeemacs import*
from bike.refactor.test_renameFunction import RenameFunctionTests, RenameFunctionTests_importsFunction
from bike.refactor.test_renameClass import RenameClassTests, RenameClassTests_importsClass


class TestLocateNode(unittest.TestCase):
    def test_findsFunctioninNestedFile(self):
        src = testdata.Function[1:]
        createPackageStructure("pass", src)
        server = Server()
        server.load("a")
        removePackageStructure()
        filename = os.path.join(os.path.abspath("a"), "b", "bah.py")
        assert server._locateNode(filename, 8, 1).fqn == "a.b.bah.theFunction"

    def test_returnsNoneIfCantFindNode(self):
        src = testdata.Function[1:]
        createPackageStructure("pass", src)
        server = Server()
        server.load("a")
        removePackageStructure()
        filename = os.path.join(os.path.abspath("a"), "b", "bah.py")
        assert server._locateNode(filename, 15, 1) == None

    def test_findsClassinNestedFile(self):
        src = testdata.RaiseClassException[1:]
        createPackageStructure("pass", src)
        server = Server()
        server.load("a")
        removePackageStructure()
        filename = os.path.join(os.path.abspath("a"), "b", "bah.py")
        assert server._locateNode(filename, 6, 1).fqn == "a.b.bah.TheClass"

class TestReloadSource(unittest.TestCase):
    def test_changesASTToReflectNewSource(self):
        createPackageStructure("pass", "pass")
        server = Server()
        server.load("a")
        removePackageStructure()
        createPackageStructure("pass", "class TheClass: pass")
        filename = os.path.join(os.path.abspath("a"), "b", "bah.py")
        server.processCommand("reload "+filename)
        removePackageStructure()
        assert server._locateNode(filename, 6,1).fqn == "a.b.bah.TheClass"

    def test_canRenameClassOnNewReloadedCode(self):
        createPackageStructure(testdata.functionContaingObjectCreationAndMethodReferenceWithImportOfNestedClass, "pass")
        server = Server()
        server.load("a")
        removePackageStructure()
        createPackageStructure("pass", "class TheClass: pass")
        filename = os.path.join(os.path.abspath("a"), "b", "bah.py")
        server.processCommand("reload "+filename)
        
        classnode = server.ast.typeinfo.getTypeOf("a.b.bah.TheClass")
        classnode.sourcenode.tokenize()
        toks = classnode.getNodeTokens()
        commandstr = "rename "+classnode.sourcenode.filename+" "+str(toks[1].start[1])+" "+str(toks[1].start[0])+" "+"NewName"
        server.processCommand(commandstr)
        removePackageStructure()
        assert getASTNode(server.ast,"CallFunc").node.attrname == "NewName"


class TestRenameFunction(RenameFunctionTests, RenameClassTests, unittest.TestCase):
    def runTarget(self, src, fqn, newname):
        try:
            writeTmpTestFile(src)
            server = Server()
            commandstr = "load "+os.path.abspath(tmpfile)
            server.processCommand(commandstr)
            fn = server.ast.typeinfo.getTypeOf(tmpmodule+"."+fqn)
            fn.sourcenode.tokenize()
            toks = fn.getNodeTokens()
            commandstr = "rename "+fn.sourcenode.filename+" "+str(toks[1].start[1])+" "+str(toks[1].start[0])+" "+newname
            server.processCommand(commandstr)
            newsrc = readFile(tmpfile)
            ast = parse(newsrc)
            return ast
        finally:
            deleteTmpTestFile()


class TestRenameMethod(unittest.TestCase):
    def runTarget(self, src, fqn, newname):
        try:
            writeTmpTestFile(src)
            server = Server()
            commandstr = "load "+os.path.abspath(tmpfile)
            server.processCommand(commandstr)
            fn = server.ast.typeinfo.getTypeOf(tmpmodule+"."+fqn)
            toks = fn.getNodeTokens()
            commandstr = "rename "+fn.sourcenode.filename+" "+str(toks[1].start[1])+" "+str(toks[1].start[0])+" "+newname
            server.processCommand(commandstr)
            newsrc = readFile(tmpfile)
            ast = Top()
            ast.addSource(Source(tmpfile,tmpmodule,newsrc))
            return ast
        finally:
            deleteTmpTestFile()

    def test_getPromptStr(self):
        src = testdata.ObjectCreatedInParentScopeAfterMethodReference
        ast = self.runTarget(src,"TheClass.theMethod","bah")
        node = getASTNode(ast,"Getattr")
        assert Server()._getPromptStr(node) == "BRMPromptForRename: "+os.path.abspath(tmpfile)+" 4 6 9 6"


class ExecuteRenameHelper:
    def executeRename(self, src, fqn, newname):
        try:
            server = Server()
            commandstr = "load "+os.path.abspath("a")
            server.processCommand(commandstr)
            fn = server.ast.typeinfo.getTypeOf(fqn)
            fn.sourcenode.tokenize()
            toks = fn.getNodeTokens()
            commandstr = "rename "+fn.sourcenode.filename+" "+str(toks[1].start[1])+" "+str(toks[1].start[0])+" "+newname
            server.processCommand(commandstr)
            newsrc = readFile(os.path.join("a", "foo.py"))
            ast = parse(newsrc)
            return ast
        finally:
            removePackageStructure()



class TestRenameFunction_importsFunction(RenameFunctionTests_importsFunction, 
                                         ExecuteRenameHelper, 
                                         unittest.TestCase):

    def runTarget(self, src, fqn, newname):
        createPackageStructure(src, testdata.Function)
        return self.executeRename(src, fqn, newname)


class TestRenameClass_importsClass(RenameClassTests_importsClass, 
                                   ExecuteRenameHelper, 
                                   unittest.TestCase):

    def runTarget(self, src, fqn, newname):
        createPackageStructure(src, testdata.TheClass)
        return self.executeRename(src, fqn, newname)

if __name__ == "__main__":
    unittest.main()
